/*
 * Copyright Contributors to the fusa-gcc-plugin project
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#include <exception>
#include <filesystem>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <optional>
#include <set>
#include <sstream>
#include <unordered_set>
#include <vector>

#include "gcc-plugin.h"
#include "plugin-version.h"

#include "context.h"
#include "diagnostic-core.h"
#include "function.h"

#include "tree.h"

#include "tree-hash-traits.h"
#include "tree-pass.h"

#include "gimple.h"

#include "gimple-expr.h"
#include "gimple-iterator.h"
#include "gimple-walk.h"

#include "input.h"

#include "candidate.hpp"
#include "cryptoutils.hpp"
#include "declhelpers.hpp"
#include "distribution.hpp"
#include "json.hpp"
#include "macrohelpers.hpp"
#include "process.hpp"
#include "fusa.hpp"

int plugin_is_GPL_compatible;

using namespace std;

static plugin_args args;

namespace {
string bold(string s) { return ("[1m" + s + "[0m"); }

string error_message_for(string name, string from_file,
                         enum api_type api_type) {
  string type;

  switch (api_type) {
  case API_TYPE_MACRO:
    type = "Macro";
    break;
  case API_TYPE_FUNCTION:
    type = "Symbol";
    break;
  }

  stringstream sb;
  sb << "[FuSa]: " << type << ": " << bold(name) << " from ("
     << from_file << ")"
     << " is not " << bold(args.profile) << " certified API" << endl;

  return sb.str();
}

const pass_data rh_fusa_pass_data = {
    GIMPLE_PASS,
    "FuSa Verified Toolchain Pass", /* name */
    OPTGROUP_NONE,                     /* optinfo_flags */
    TV_NONE,                           /* tv_id */
    PROP_gimple_any,                   /* properties_required */
    0,                                 /* properties_provided */
    0,                                 /* properties_destroyed */
    0,                                 /* todo_flags_start */
    0                                  /* todo_flags_finish */
};

struct fusa_pass : gimple_opt_pass {
  size_t num_errors;
  candidate::set candidates;
  vector<certified_symbol> symbols;
  plugin_args args;

  fusa_pass(gcc::context *ctx, plugin_args args,
               vector<certified_symbol> symbols)
      : gimple_opt_pass(rh_fusa_pass_data, ctx) {
    this->symbols = symbols;
    this->args = args;
    this->num_errors = 0;
  }

  candidate::set *extract_candidates(::function *f) {
    candidate::collection candidates;

    struct walk_stmt_info walk_stmt_info = {};
    walk_stmt_info.info = &candidates;

    walk_gimple_seq(
        f->gimple_body,
        [](gimple_stmt_iterator *gsi, bool *handled_all_ops,
           struct walk_stmt_info *wi) -> tree {
          candidate::collection *c =
              static_cast<candidate::collection *>(wi->info);

          gimple *g = gsi_stmt(*gsi);
          enum gimple_code code = gimple_code(g);
          location_t location = gimple_location(g);

          if (optional<candidate> c_macro = candidate::from_location(location))
            c->items->insert(*c_macro);

          switch (code) {
          case GIMPLE_ASSIGN: {
            c->qualify_tree_recursive(gimple_assign_rhs1(g), location);
            c->qualify_tree_recursive(gimple_assign_rhs2(g), location);
            c->qualify_tree_recursive(gimple_assign_rhs3(g), location);
            break;
          }

          case GIMPLE_CALL: {
            c->qualify_tree_recursive(gimple_call_fndecl(g), location);

            unsigned int nargs = gimple_call_num_args(g);
            for (unsigned int i = 0; i < nargs; ++i)
              c->qualify_tree_recursive(gimple_call_arg(g, i), location);

            break;
          }

          case GIMPLE_RETURN: {
            c->qualify_tree_recursive(
                gimple_return_retval(static_cast<const greturn *>(g)),
                location);
            break;
          }

          case GIMPLE_COND: {
            c->qualify_tree_recursive(gimple_cond_lhs(g), location);
            c->qualify_tree_recursive(gimple_cond_rhs(g), location);
            break;
          }

          default:
            /**
             * What other Gimple codes should we handle?
             * cerr << "Code: " << gimple_code_name[code] << endl;
             */
            break;
          }

          return NULL;
        },
        NULL, &walk_stmt_info);

    return candidates.items;
  }

  virtual unsigned int execute(::function *f) override {
    /**
     * Check if the top level decl within the current CU comes from
     * path owned by distribution - this happens when code gets inlined
     * from header files to the current CU.
     **/
    if (optional<candidate> c = candidate::from_decl(f->decl, 0)) {
      if (!c->fusa_check(this->symbols, this->candidates, this->args)) {

        string message =
            error_message_for(c->name, c->decl_source_file, c->api_type);

        error_at(f->function_start_locus, message.c_str());

        this->num_errors += 1;
      }
    } else {
      candidate::set *to_check = this->extract_candidates(f);

      for (candidate const &candidate : *to_check) {
        if (!candidate.fusa_check(this->symbols, this->candidates,
                                  this->args)) {

          string message = error_message_for(
              candidate.name, candidate.decl_source_file, candidate.api_type);

          error_at(candidate.invocation_location, message.c_str());

          this->num_errors += 1;
        }
      }

      delete to_check;
    }

    return 0;
  }

  virtual fusa_pass *clone() override { return this; }
};

void pass_finish(void *gcc_data, void *user_data) {
  if (!args.dump_apis)
    return;

  const char *input_filename = LOCATION_FILE(input_location);
  if (!input_filename) {
    inform((location_t)0, "Unknown input path");
    return;
  }

  filesystem::path base = args.dump_path;
  if (!filesystem::exists(base)) {
    error("[FuSa] Path for storing output doesn't exist: %s",
          base.c_str());
    return;
  }

  filesystem::path input_path = filesystem::absolute(input_filename);
  filesystem::path final_path =
      base / ("qualify_" + crypto::sha256_string(input_path) + ".json");

  candidate::set *all_apis = static_cast<candidate::set *>(user_data);

  string datastr;
  if (serialize_output(*all_apis, datastr) != 0) {
    error("Failed serializing output JSON");
    return;
  }

  ofstream dest(final_path);

  if (dest.fail()) {
    error("Couldn't write to: %s", final_path.c_str());
    return;
  }

  dest << datastr;

  inform((location_t)0, "[FuSa] Symbol database created at: %s", final_path.c_str());
}

} // namespace

namespace arguments {
struct arg_def {
  string name;
  string description;

  arg_def(string name, string description)
      : name(name), description(description) {}

  virtual void ingest(plugin_args *args, const char *value) = 0;

  bool match_arg(const char *s) {
    return strncmp(s, this->name.c_str(), this->name.size()) == 0;
  }
};

struct dump_apis_arg_def : arg_def {
  dump_apis_arg_def()
      : arg_def("dump-apis", "Dump symbol database to the specified file") {}

  virtual void ingest(plugin_args *args, const char *value) override {
    args->dump_apis = true;
    args->dump_path = value;
  }
};

const static unordered_set<string> valid_profiles = {"QM", "ASIL-B"};
struct target_profile_arg_def : arg_def {

  target_profile_arg_def()
      : arg_def("target-profile",
                "Target profile for safety requirements. Valid options: "
                "asil-b, qm. Default: asil-b") {}

  virtual void ingest(plugin_args *args, const char *value) override {
    string profile = value;
    for (auto &c : profile)
      c = TOUPPER(c);

    if (valid_profiles.find(profile) == valid_profiles.end()) {
      error("[FuSa] Unknown profile: %s", value);
    }

    args->profile = profile;
  }
};

struct database_arg_def : arg_def {
  database_arg_def()
      : arg_def("database-file", "Use symbol database from another location") {}

  virtual void ingest(plugin_args *args, const char *value) override {
    args->database_file = value;
  }
};

struct print_usage_arg_def : arg_def {
  print_usage_arg_def() : arg_def("usage", "Print usage and exit") {}

  virtual void ingest(plugin_args *args, const char *value) override {
    args->print_usage = true;
  }
};

vector<arg_def *> arg_defs = {new dump_apis_arg_def(), new database_arg_def(),
                              new target_profile_arg_def(),
                              new print_usage_arg_def()};

plugin_args parse_args(struct plugin_name_args *plugin_info) {
  plugin_args args = {};

  const plugin_argument *argv = plugin_info->argv;

  for (int i = 0; i < plugin_info->argc; ++i) {
    for (auto &def : arg_defs) {
      if (def->match_arg(argv[i].key)) {
        def->ingest(&args, argv[i].value);
      }
    }
  }

  return args;
}
} // namespace arguments

struct plugin_info make_plugin_info() {
  stringstream ss;

  ss << "FuSa Toolchain" << endl
     << endl
     << "This plugin verifies that only FuSa certified functions are used "
        "during "
        "compilation."
     << endl
     << endl
     << "Arguments:" << endl
     << endl;

  for (auto &def : arguments::arg_defs) {
    string arg_name = "  -fplugin-arg-fusa-";
    arg_name.append(def->name);

    ss << setw(45) << left << arg_name << " - " << def->description << endl;
  }

  string *description = new string(ss.str());
  return {"0.1", description->c_str()}; // memory leak ok
}

int plugin_init(struct plugin_name_args *plugin_info,
                struct plugin_gcc_version *version) {
  // This version check is very strict. We have decided not to validate version for now
  // if (!plugin_default_version_check(version, &gcc_version)) {
  //   error("FuSa GCC plugin is for GCC version %u.%u\n", GCCPLUGIN_VERSION_MAJOR, GCCPLUGIN_VERSION_MINOR);
  //   return 1;
  // }

  struct plugin_info info = make_plugin_info();

  args = arguments::parse_args(plugin_info);

  if (args.print_usage) {
    cerr << info.help << endl;
    return 0;
  }

  vector<certified_symbol> symbols;
  if (!args.dump_apis) {
    int r = load_symbols_from_file(args.database_file, symbols);

    if (r != 0) {
      error("[FuSa] Error loading certified symbols from: %s\n", args.database_file.c_str());
      return r;
    }
  }

  fusa_pass *pass = new fusa_pass(g, args, symbols);

  struct register_pass_info pass_info;
  pass_info.pass = pass;
  pass_info.reference_pass_name = "cfg";
  pass_info.ref_pass_instance_number = 1;
  pass_info.pos_op = PASS_POS_INSERT_BEFORE;

  register_callback(plugin_info->base_name, PLUGIN_INFO, NULL, &info);
  register_callback(plugin_info->base_name, PLUGIN_PASS_MANAGER_SETUP, NULL,
                    &pass_info);
  register_callback(plugin_info->base_name, PLUGIN_FINISH, pass_finish,
                    &pass->candidates);

  return 0;
}
